@@ -25,7 +25,7 @@
 # OTHER DEALINGS IN THE SOFTWARE.
 from __future__ import unicode_literals
 
-__version__ = version = '0.6.4'
+__version__ = version = '0.7.0'
 
 import sys
 
@@ -38,8 +38,10 @@ activate = httprettified
 
 SELF = sys.modules[__name__]
 
-for attr in [name.decode() for name in httpretty.METHODS] + ['register_uri', 'enable', 'disable', 'is_enabled', 'Response']:
+for attr in [name.decode() for name in httpretty.METHODS] + ['register_uri', 'enable', 'disable', 'is_enabled', 'reset', 'Response']:
     setattr(SELF, attr, getattr(httpretty, attr))
 
+
 def last_request():
+    """returns the last request"""
     return httpretty.last_request
